@@ -24,15 +24,17 @@
 #include "tok_spec_struct.h"
 #include "trace.h"
 
-pthread_rwlock_t sess_list_rwlock = PTHREAD_RWLOCK_INITIALIZER;
 // session_mgr_find()
 //
 // search for the specified session. returning a pointer to the session
-// might be dangerous, but performs well
+// might be dangerous, but performs well.
+//
+// The returned session must be put back (using bt_put_node_value()) by the
+// caller to decrease the reference count!
 //
 // Returns:  SESSION * or NULL
 //
-SESSION *session_mgr_find(CK_SESSION_HANDLE handle)
+SESSION *session_mgr_find(STDLL_TokData_t *tokdata, CK_SESSION_HANDLE handle)
 {
     SESSION *result = NULL;
 
@@ -40,11 +42,15 @@ SESSION *session_mgr_find(CK_SESSION_HANDLE handle)
         return NULL;
     }
 
-    result = bt_get_node_value(&sess_btree, handle);
+    result = bt_get_node_value(&tokdata->sess_btree, handle);
 
     return result;
 }
 
+void session_mgr_put(STDLL_TokData_t *tokdata, SESSION *session)
+{
+    bt_put_node_value(&tokdata->sess_btree, session);
+}
 
 // session_mgr_new()
 //
@@ -56,8 +62,8 @@ SESSION *session_mgr_find(CK_SESSION_HANDLE handle)
 //
 // Returns:  CK_RV
 //
-CK_RV session_mgr_new(CK_ULONG flags, CK_SLOT_ID slot_id,
-                      CK_SESSION_HANDLE_PTR phSession)
+CK_RV session_mgr_new(STDLL_TokData_t *tokdata, CK_ULONG flags,
+                      CK_SLOT_ID slot_id, CK_SESSION_HANDLE_PTR phSession)
 {
     SESSION *new_session = NULL;
     CK_BBOOL user_session = FALSE;
@@ -85,8 +91,14 @@ CK_RV session_mgr_new(CK_ULONG flags, CK_SLOT_ID slot_id,
     // that all sessions belonging to a process have the same login/logout
     // status
     //
-    so_session = session_mgr_so_session_exists();
-    user_session = session_mgr_user_session_exists();
+    so_session = session_mgr_so_session_exists(tokdata);
+    user_session = session_mgr_user_session_exists(tokdata);
+
+    if (pthread_rwlock_wrlock(&tokdata->sess_list_rwlock)) {
+        TRACE_ERROR("Write Lock failed.\n");
+        rc = CKR_CANT_LOCK;
+        goto done;
+    }
 
     // we don't have to worry about having a user and SO session at the same
     // time. that is prevented in the login routine
@@ -96,7 +108,7 @@ CK_RV session_mgr_new(CK_ULONG flags, CK_SLOT_ID slot_id,
             new_session->session_info.state = CKS_RW_USER_FUNCTIONS;
         } else {
             new_session->session_info.state = CKS_RO_USER_FUNCTIONS;
-            ro_session_count++;
+            tokdata->ro_session_count++;
         }
     } else if (so_session) {
         new_session->session_info.state = CKS_RW_SO_FUNCTIONS;
@@ -105,11 +117,13 @@ CK_RV session_mgr_new(CK_ULONG flags, CK_SLOT_ID slot_id,
             new_session->session_info.state = CKS_RW_PUBLIC_SESSION;
         } else {
             new_session->session_info.state = CKS_RO_PUBLIC_SESSION;
-            ro_session_count++;
+            tokdata->ro_session_count++;
         }
     }
 
-    *phSession = bt_node_add(&sess_btree, new_session);
+    pthread_rwlock_unlock(&tokdata->sess_list_rwlock);
+
+    *phSession = bt_node_add(&tokdata->sess_btree, new_session);
     if (*phSession == 0) {
         rc = CKR_HOST_MEMORY;
         /* new_session will be free'd below */
@@ -131,18 +145,18 @@ done:
 //
 // Returns:  TRUE or FALSE
 //
-CK_BBOOL session_mgr_so_session_exists(void)
+CK_BBOOL session_mgr_so_session_exists(STDLL_TokData_t *tokdata)
 {
     CK_BBOOL result;
 
     /* we must acquire sess_list_rwlock in order to inspect
      * global_login_state */
-    if (pthread_rwlock_rdlock(&sess_list_rwlock)) {
+    if (pthread_rwlock_rdlock(&tokdata->sess_list_rwlock)) {
         TRACE_ERROR("Read Lock failed.\n");
         return FALSE;
     }
-    result = (global_login_state == CKS_RW_SO_FUNCTIONS);
-    pthread_rwlock_unlock(&sess_list_rwlock);
+    result = (tokdata->global_login_state == CKS_RW_SO_FUNCTIONS);
+    pthread_rwlock_unlock(&tokdata->sess_list_rwlock);
 
     return result;
 }
@@ -154,20 +168,20 @@ CK_BBOOL session_mgr_so_session_exists(void)
 //
 // Returns:  TRUE or FALSE
 //
-CK_BBOOL session_mgr_user_session_exists(void)
+CK_BBOOL session_mgr_user_session_exists(STDLL_TokData_t *tokdata)
 {
     CK_BBOOL result;
 
     /* we must acquire sess_list_rwlock in order to inspect
      * glogal_login_state */
-    if (pthread_rwlock_rdlock(&sess_list_rwlock)) {
+    if (pthread_rwlock_rdlock(&tokdata->sess_list_rwlock)) {
         TRACE_ERROR("Read Lock failed.\n");
         return FALSE;
     }
-    result = ((global_login_state == CKS_RO_USER_FUNCTIONS) ||
-              (global_login_state == CKS_RW_USER_FUNCTIONS));
+    result = ((tokdata->global_login_state == CKS_RO_USER_FUNCTIONS) ||
+              (tokdata->global_login_state == CKS_RW_USER_FUNCTIONS));
 
-    pthread_rwlock_unlock(&sess_list_rwlock);
+    pthread_rwlock_unlock(&tokdata->sess_list_rwlock);
 
     return result;
 }
@@ -179,20 +193,20 @@ CK_BBOOL session_mgr_user_session_exists(void)
 //
 // Returns:  TRUE or FALSE
 //
-CK_BBOOL session_mgr_public_session_exists(void)
+CK_BBOOL session_mgr_public_session_exists(STDLL_TokData_t *tokdata)
 {
     CK_BBOOL result;
 
     /* we must acquire sess_list_rwlock in order to inspect
      * global_login_state */
-    if (pthread_rwlock_rdlock(&sess_list_rwlock)) {
+    if (pthread_rwlock_rdlock(&tokdata->sess_list_rwlock)) {
         TRACE_ERROR("Read Lock failed.\n");
         return FALSE;
     }
-    result = ((global_login_state == CKS_RO_PUBLIC_SESSION) ||
-              (global_login_state == CKS_RW_PUBLIC_SESSION));
+    result = ((tokdata->global_login_state == CKS_RO_PUBLIC_SESSION) ||
+              (tokdata->global_login_state == CKS_RW_PUBLIC_SESSION));
 
-    pthread_rwlock_unlock(&sess_list_rwlock);
+    pthread_rwlock_unlock(&tokdata->sess_list_rwlock);
 
     return result;
 }
@@ -203,19 +217,19 @@ CK_BBOOL session_mgr_public_session_exists(void)
 // determines whether the specified process owns any read-only sessions. this is
 // useful because the SO cannot log in if a read-only session exists.
 //
-CK_BBOOL session_mgr_readonly_session_exists(void)
+CK_BBOOL session_mgr_readonly_session_exists(STDLL_TokData_t *tokdata)
 {
     CK_BBOOL result;
 
     /* we must acquire sess_list_rwlock in order to inspect ro_session_count */
-    if (pthread_rwlock_rdlock(&sess_list_rwlock)) {
+    if (pthread_rwlock_rdlock(&tokdata->sess_list_rwlock)) {
         TRACE_ERROR("Read Lock failed.\n");
         return FALSE;
     }
 
-    result = (ro_session_count > 0);
+    result = (tokdata->ro_session_count > 0);
 
-    pthread_rwlock_unlock(&sess_list_rwlock);
+    pthread_rwlock_unlock(&tokdata->sess_list_rwlock);
 
     return result;
 }
@@ -236,22 +250,24 @@ CK_RV session_mgr_close_session(STDLL_TokData_t *tokdata,
     SESSION *sess;
     CK_RV rc = CKR_OK;
 
-    sess = bt_get_node_value(&sess_btree, handle);
+    sess = bt_get_node_value(&tokdata->sess_btree, handle);
     if (!sess) {
         TRACE_ERROR("%s\n", ock_err(ERR_SESSION_HANDLE_INVALID));
         return CKR_SESSION_HANDLE_INVALID;
     }
 
-    if (pthread_rwlock_wrlock(&sess_list_rwlock)) {
-        TRACE_ERROR("Read Lock failed.\n");
-        return FALSE;
+    if (pthread_rwlock_wrlock(&tokdata->sess_list_rwlock)) {
+        TRACE_ERROR("Write Lock failed.\n");
+        bt_put_node_value(&tokdata->sess_btree, sess);
+        sess = NULL;
+        return CKR_CANT_LOCK;
     }
 
     object_mgr_purge_session_objects(tokdata, sess, ALL);
 
     if ((sess->session_info.state == CKS_RO_PUBLIC_SESSION) ||
         (sess->session_info.state == CKS_RO_USER_FUNCTIONS)) {
-        ro_session_count--;
+        tokdata->ro_session_count--;
     }
 
     // Make sure this address is now invalid
@@ -290,7 +306,9 @@ CK_RV session_mgr_close_session(STDLL_TokData_t *tokdata,
     if (sess->verify_ctx.mech.pParameter)
         free(sess->verify_ctx.mech.pParameter);
 
-    bt_node_free(&sess_btree, handle, free);
+    bt_put_node_value(&tokdata->sess_btree, sess);
+    sess = NULL;
+    bt_node_free(&tokdata->sess_btree, handle, TRUE);
 
     // XXX XXX  Not having this is a problem
     //  for IHS.  The spec states that there is an implicit logout
@@ -301,21 +319,21 @@ CK_RV session_mgr_close_session(STDLL_TokData_t *tokdata,
     //  objects EVERY time.   If we are logged out, we MUST purge the private
     //  objects from this process..
     //
-    if (bt_is_empty(&sess_btree)) {
+    if (bt_is_empty(&tokdata->sess_btree)) {
         // SAB  XXX  if all sessions are closed.  Is this effectivly logging out
         if (token_specific.t_logout) {
-            rc = token_specific.t_logout();
+            rc = token_specific.t_logout(tokdata);
         }
         object_mgr_purge_private_token_objects(tokdata);
 
-        global_login_state = CKS_RO_PUBLIC_SESSION;
+        tokdata->global_login_state = CKS_RO_PUBLIC_SESSION;
         // The objects really need to be purged .. but this impacts the
         // performance under linux.   So we need to make sure that the
         // login state is valid.    I don't really like this.
         object_mgr_purge_map(tokdata, (SESSION *) 0xFFFF, PRIVATE);
     }
 
-    pthread_rwlock_unlock(&sess_list_rwlock);
+    pthread_rwlock_unlock(&tokdata->sess_list_rwlock);
     return rc;
 }
 
@@ -367,26 +385,28 @@ void session_free(STDLL_TokData_t *tokdata, void *node_value,
         free(sess->verify_ctx.mech.pParameter);
 
     /* NB: any access to sess or @node_value after this returns will segfault */
-    bt_node_free(&sess_btree, node_idx, free);
+    bt_node_free(&tokdata->sess_btree, node_idx, TRUE);
 }
 
 // session_mgr_close_all_sessions()
 //
-// removes all sessions from the specified process
+// removes all sessions from the specified process.
+// If tokdata is not NULL, then only sessions for that token instance are
+// removed.
 //
-CK_RV session_mgr_close_all_sessions(void)
+CK_RV session_mgr_close_all_sessions(STDLL_TokData_t *tokdata)
 {
-    bt_for_each_node(NULL, &sess_btree, session_free, NULL);
+    bt_for_each_node(tokdata, &tokdata->sess_btree, session_free, NULL);
 
-    if (pthread_rwlock_wrlock(&sess_list_rwlock)) {
-        TRACE_ERROR("Read Lock failed.\n");
-        return FALSE;
+    if (pthread_rwlock_wrlock(&tokdata->sess_list_rwlock)) {
+        TRACE_ERROR("Write Lock failed.\n");
+        return CKR_CANT_LOCK;
     }
 
-    global_login_state = CKS_RO_PUBLIC_SESSION;
-    ro_session_count = 0;
+    tokdata->global_login_state = CKS_RO_PUBLIC_SESSION;
+    tokdata->ro_session_count = 0;
 
-    pthread_rwlock_unlock(&sess_list_rwlock);
+    pthread_rwlock_unlock(&tokdata->sess_list_rwlock);
 
     return CKR_OK;
 }
@@ -415,7 +435,7 @@ void session_login(STDLL_TokData_t *tokdata, void *node_value,
             s->session_info.state = CKS_RO_USER_FUNCTIONS;
     }
 
-    global_login_state = s->session_info.state; // SAB
+    tokdata->global_login_state = s->session_info.state; // SAB
 }
 
 // session_mgr_login_all()
@@ -426,7 +446,15 @@ void session_login(STDLL_TokData_t *tokdata, void *node_value,
 //
 CK_RV session_mgr_login_all(STDLL_TokData_t *tokdata, CK_USER_TYPE user_type)
 {
-    bt_for_each_node(tokdata, &sess_btree, session_login, (void *) &user_type);
+    if (pthread_rwlock_wrlock(&tokdata->sess_list_rwlock)) {
+        TRACE_ERROR("Write Lock failed.\n");
+        return CKR_CANT_LOCK;
+    }
+
+    bt_for_each_node(tokdata, &tokdata->sess_btree, session_login,
+                     (void *)&user_type);
+
+    pthread_rwlock_unlock(&tokdata->sess_list_rwlock);
 
     return CKR_OK;
 }
@@ -453,7 +481,7 @@ void session_logout(STDLL_TokData_t *tokdata, void *node_value,
     else
         s->session_info.state = CKS_RO_PUBLIC_SESSION;
 
-    global_login_state = s->session_info.state; // SAB
+    tokdata->global_login_state = s->session_info.state; // SAB
 }
 
 // session_mgr_logout_all()
@@ -462,7 +490,14 @@ void session_logout(STDLL_TokData_t *tokdata, void *node_value,
 //
 CK_RV session_mgr_logout_all(STDLL_TokData_t *tokdata)
 {
-    bt_for_each_node(tokdata, &sess_btree, session_logout, NULL);
+    if (pthread_rwlock_wrlock(&tokdata->sess_list_rwlock)) {
+        TRACE_ERROR("Write Lock failed.\n");
+        return CKR_CANT_LOCK;
+    }
+
+    bt_for_each_node(tokdata, &tokdata->sess_btree, session_logout, NULL);
+
+    pthread_rwlock_unlock(&tokdata->sess_list_rwlock);
 
     return CKR_OK;
 }
@@ -505,6 +540,11 @@ CK_RV session_mgr_get_op_state(SESSION *sess,
         if (length_only == FALSE) {
             op_data = (OP_STATE_DATA *) data;
 
+            if (*data_len < op_data_len) {
+                TRACE_ERROR("%s\n", ock_err(ERR_BUFFER_TOO_SMALL));
+                 return CKR_BUFFER_TOO_SMALL;
+            }
+
             op_data->data_len = op_data_len - sizeof(OP_STATE_DATA);
             op_data->session_state = sess->session_info.state;
             op_data->active_operation = STATE_ENCR;
@@ -544,6 +584,11 @@ CK_RV session_mgr_get_op_state(SESSION *sess,
         if (length_only == FALSE) {
             op_data = (OP_STATE_DATA *) data;
 
+            if (*data_len < op_data_len) {
+                TRACE_ERROR("%s\n", ock_err(ERR_BUFFER_TOO_SMALL));
+                 return CKR_BUFFER_TOO_SMALL;
+            }
+
             op_data->data_len = op_data_len - sizeof(OP_STATE_DATA);
             op_data->session_state = sess->session_info.state;
             op_data->active_operation = STATE_DECR;
@@ -583,6 +628,11 @@ CK_RV session_mgr_get_op_state(SESSION *sess,
         if (length_only == FALSE) {
             op_data = (OP_STATE_DATA *) data;
 
+            if (*data_len < op_data_len) {
+                TRACE_ERROR("%s\n", ock_err(ERR_BUFFER_TOO_SMALL));
+                 return CKR_BUFFER_TOO_SMALL;
+            }
+
             op_data->data_len = op_data_len - sizeof(OP_STATE_DATA);
             op_data->session_state = sess->session_info.state;
             op_data->active_operation = STATE_DIGEST;
@@ -622,6 +672,11 @@ CK_RV session_mgr_get_op_state(SESSION *sess,
         if (length_only == FALSE) {
             op_data = (OP_STATE_DATA *) data;
 
+            if (*data_len < op_data_len) {
+                TRACE_ERROR("%s\n", ock_err(ERR_BUFFER_TOO_SMALL));
+                 return CKR_BUFFER_TOO_SMALL;
+            }
+
             op_data->data_len = op_data_len - sizeof(OP_STATE_DATA);
             op_data->session_state = sess->session_info.state;
             op_data->active_operation = STATE_SIGN;
@@ -661,9 +716,14 @@ CK_RV session_mgr_get_op_state(SESSION *sess,
         if (length_only == FALSE) {
             op_data = (OP_STATE_DATA *) data;
 
+            if (*data_len < op_data_len) {
+                TRACE_ERROR("%s\n", ock_err(ERR_BUFFER_TOO_SMALL));
+                 return CKR_BUFFER_TOO_SMALL;
+            }
+
             op_data->data_len = op_data_len - sizeof(OP_STATE_DATA);
             op_data->session_state = sess->session_info.state;
-            op_data->active_operation = STATE_SIGN;
+            op_data->active_operation = STATE_VERIFY;
 
             offset = sizeof(OP_STATE_DATA);
 
@@ -712,14 +772,17 @@ CK_RV session_mgr_set_op_state(SESSION *sess,
     CK_BYTE *ptr3 = NULL;
     CK_ULONG len;
 
-    UNUSED(data_len);
-
     if (!sess || !data) {
         TRACE_ERROR("%s received bad argument(s)\n", __func__);
         return CKR_FUNCTION_FAILED;
     }
     op_data = (OP_STATE_DATA *) data;
 
+    if (data_len < op_data->data_len + sizeof(OP_STATE_DATA)) {
+        TRACE_ERROR("%s\n", ock_err(ERR_SAVED_STATE_INVALID));
+        return CKR_SAVED_STATE_INVALID;
+    }
+
     // make sure the session states are compatible
     //
     if (sess->session_info.state != op_data->session_state) {
@@ -743,11 +806,11 @@ CK_RV session_mgr_set_op_state(SESSION *sess,
                 TRACE_ERROR("%s\n", ock_err(ERR_SAVED_STATE_INVALID));
                 return CKR_SAVED_STATE_INVALID;
             }
-            if (auth_key != 0) {
+            if (auth_key != CK_INVALID_HANDLE) {
                 TRACE_ERROR("%s\n", ock_err(ERR_KEY_NOT_NEEDED));
                 return CKR_KEY_NOT_NEEDED;
             }
-            if (encr_key == 0) {
+            if (encr_key == CK_INVALID_HANDLE) {
                 TRACE_ERROR("%s\n", ock_err(ERR_KEY_NEEDED));
                 return CKR_KEY_NEEDED;
             }
@@ -789,11 +852,11 @@ CK_RV session_mgr_set_op_state(SESSION *sess,
                 TRACE_ERROR("%s\n", ock_err(ERR_SAVED_STATE_INVALID));
                 return CKR_SAVED_STATE_INVALID;
             }
-            if (auth_key == 0) {
+            if (auth_key == CK_INVALID_HANDLE) {
                 TRACE_ERROR("%s\n", ock_err(ERR_KEY_NEEDED));
                 return CKR_KEY_NEEDED;
             }
-            if (encr_key != 0) {
+            if (encr_key != CK_INVALID_HANDLE) {
                 TRACE_ERROR("%s\n", ock_err(ERR_KEY_NOT_NEEDED));
                 return CKR_KEY_NOT_NEEDED;
             }
@@ -834,11 +897,11 @@ CK_RV session_mgr_set_op_state(SESSION *sess,
                 TRACE_ERROR("%s\n", ock_err(ERR_SAVED_STATE_INVALID));
                 return CKR_SAVED_STATE_INVALID;
             }
-            if (auth_key != 0) {
+            if (auth_key != CK_INVALID_HANDLE) {
                 TRACE_ERROR("%s\n", ock_err(ERR_KEY_NOT_NEEDED));
                 return CKR_KEY_NOT_NEEDED;
             }
-            if (encr_key != 0) {
+            if (encr_key != CK_INVALID_HANDLE) {
                 TRACE_ERROR("%s\n", ock_err(ERR_KEY_NOT_NEEDED));
                 return CKR_KEY_NOT_NEEDED;
             }
