@@ -50,7 +50,7 @@ CK_RV pk_des_ecb_encrypt(STDLL_TokData_t *tokdata,
         TRACE_ERROR("%s\n", ock_err(ERR_DATA_LEN_RANGE));
         return CKR_DATA_LEN_RANGE;
     }
-    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key);
+    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key, READ_LOCK);
     if (rc != CKR_OK) {
         TRACE_ERROR("Failed to find specified object.\n");
         return rc;
@@ -58,17 +58,25 @@ CK_RV pk_des_ecb_encrypt(STDLL_TokData_t *tokdata,
 
     if (length_only == TRUE) {
         *out_data_len = in_data_len;
-        return CKR_OK;
+        rc = CKR_OK;
+        goto done;
     }
 
     if (*out_data_len < in_data_len) {
         *out_data_len = in_data_len;
         TRACE_ERROR("%s\n", ock_err(ERR_BUFFER_TOO_SMALL));
-        return CKR_BUFFER_TOO_SMALL;
+        rc = CKR_BUFFER_TOO_SMALL;
+        goto done;
     }
 
-    return ckm_des_ecb_encrypt(tokdata, in_data, in_data_len,
-                               out_data, out_data_len, key);
+    rc = ckm_des_ecb_encrypt(tokdata, in_data, in_data_len,
+                             out_data, out_data_len, key);
+
+done:
+    object_put(tokdata, key, TRUE);
+    key = NULL;
+
+    return rc;
 }
 
 
@@ -97,7 +105,7 @@ CK_RV des_ecb_decrypt(STDLL_TokData_t *tokdata,
         TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_LEN_RANGE));
         return CKR_ENCRYPTED_DATA_LEN_RANGE;
     }
-    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key);
+    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key, READ_LOCK);
     if (rc != CKR_OK) {
         TRACE_ERROR("Failed to find specified object.\n");
         return rc;
@@ -105,17 +113,25 @@ CK_RV des_ecb_decrypt(STDLL_TokData_t *tokdata,
 
     if (length_only == TRUE) {
         *out_data_len = in_data_len;
-        return CKR_OK;
+        rc = CKR_OK;
+        goto done;
     }
 
     if (*out_data_len < in_data_len) {
         *out_data_len = in_data_len;
         TRACE_ERROR("%s\n", ock_err(ERR_BUFFER_TOO_SMALL));
-        return CKR_BUFFER_TOO_SMALL;
+        rc = CKR_BUFFER_TOO_SMALL;
+        goto done;
     }
 
-    return ckm_des_ecb_decrypt(tokdata, in_data, in_data_len,
-                               out_data, out_data_len, key);
+    rc = ckm_des_ecb_decrypt(tokdata, in_data, in_data_len,
+                             out_data, out_data_len, key);
+
+done:
+    object_put(tokdata, key, TRUE);
+    key = NULL;
+
+    return rc;
 }
 
 
@@ -143,7 +159,7 @@ CK_RV pk_des_cbc_encrypt(STDLL_TokData_t *tokdata,
         TRACE_ERROR("%s\n", ock_err(ERR_DATA_LEN_RANGE));
         return CKR_DATA_LEN_RANGE;
     }
-    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key);
+    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key, READ_LOCK);
     if (rc != CKR_OK) {
         TRACE_ERROR("Failed to find specified object.\n");
         return rc;
@@ -151,17 +167,25 @@ CK_RV pk_des_cbc_encrypt(STDLL_TokData_t *tokdata,
 
     if (length_only == TRUE) {
         *out_data_len = in_data_len;
-        return CKR_OK;
+        rc = CKR_OK;
+        goto done;
     }
 
     if (*out_data_len < in_data_len) {
         *out_data_len = in_data_len;
         TRACE_ERROR("%s\n", ock_err(ERR_BUFFER_TOO_SMALL));
-        return CKR_BUFFER_TOO_SMALL;
+        rc = CKR_BUFFER_TOO_SMALL;
+        goto done;
     }
 
-    return ckm_des_cbc_encrypt(tokdata, in_data, in_data_len, out_data,
-                               out_data_len, ctx->mech.pParameter, key);
+    rc = ckm_des_cbc_encrypt(tokdata, in_data, in_data_len, out_data,
+                             out_data_len, ctx->mech.pParameter, key);
+
+done:
+    object_put(tokdata, key, TRUE);
+    key = NULL;
+
+    return rc;
 }
 
 
@@ -190,7 +214,7 @@ CK_RV des_cbc_decrypt(STDLL_TokData_t *tokdata,
         TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_LEN_RANGE));
         return CKR_ENCRYPTED_DATA_LEN_RANGE;
     }
-    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key);
+    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key, READ_LOCK);
     if (rc != CKR_OK) {
         TRACE_ERROR("Failed to find specified object.\n");
         return rc;
@@ -198,17 +222,25 @@ CK_RV des_cbc_decrypt(STDLL_TokData_t *tokdata,
 
     if (length_only == TRUE) {
         *out_data_len = in_data_len;
-        return CKR_OK;
+        rc = CKR_OK;
+        goto done;
     }
 
     if (*out_data_len < in_data_len) {
         *out_data_len = in_data_len;
         TRACE_ERROR("%s\n", ock_err(ERR_BUFFER_TOO_SMALL));
-        return CKR_BUFFER_TOO_SMALL;
+        rc = CKR_BUFFER_TOO_SMALL;
+        goto done;
     }
 
-    return ckm_des_cbc_decrypt(tokdata, in_data, in_data_len, out_data,
-                               out_data_len, ctx->mech.pParameter, key);
+    rc = ckm_des_cbc_decrypt(tokdata, in_data, in_data_len, out_data,
+                             out_data_len, ctx->mech.pParameter, key);
+
+done:
+    object_put(tokdata, key, TRUE);
+    key = NULL;
+
+    return rc;
 }
 
 
@@ -234,7 +266,7 @@ CK_RV des_cbc_pad_encrypt(STDLL_TokData_t *tokdata,
     // DES-CBC-PAD has no input length requirements
     //
 
-    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key);
+    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key, READ_LOCK);
     if (rc != CKR_OK) {
         TRACE_ERROR("Failed to find specified object.\n");
         return rc;
@@ -245,21 +277,25 @@ CK_RV des_cbc_pad_encrypt(STDLL_TokData_t *tokdata,
 
     if (length_only == TRUE) {
         *out_data_len = padded_len;
-        return CKR_OK;
+        rc = CKR_OK;
+        goto done;
     }
 
     if (*out_data_len < padded_len) {
         *out_data_len = padded_len;
         TRACE_ERROR("%s\n", ock_err(ERR_BUFFER_TOO_SMALL));
-        return CKR_BUFFER_TOO_SMALL;
+        rc = CKR_BUFFER_TOO_SMALL;
+        goto done;
     }
 
     clear = (CK_BYTE *) malloc(padded_len);
     if (!clear) {
         TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
-        return CKR_HOST_MEMORY;
+        rc = CKR_HOST_MEMORY;
+        goto done;
     }
-    memcpy(clear, in_data, in_data_len);
+    if (in_data != NULL && in_data_len > 0)
+        memcpy(clear, in_data, in_data_len);
 
     add_pkcs_padding(clear + in_data_len,
                      DES_BLOCK_SIZE, in_data_len, padded_len);
@@ -268,6 +304,10 @@ CK_RV des_cbc_pad_encrypt(STDLL_TokData_t *tokdata,
                              ctx->mech.pParameter, key);
     free(clear);
 
+done:
+    object_put(tokdata, key, TRUE);
+    key = NULL;
+
     return rc;
 }
 
@@ -296,7 +336,7 @@ CK_RV des_cbc_pad_decrypt(STDLL_TokData_t *tokdata,
     // no need to validate the input length since we'll pad as necessary
     //
 
-    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key);
+    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key, READ_LOCK);
     if (rc != CKR_OK) {
         TRACE_ERROR("Failed to find specified object.\n");
         return rc;
@@ -306,7 +346,8 @@ CK_RV des_cbc_pad_decrypt(STDLL_TokData_t *tokdata,
     //
     if (in_data_len % DES_BLOCK_SIZE != 0) {
         TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_LEN_RANGE));
-        return CKR_ENCRYPTED_DATA_LEN_RANGE;
+        rc = CKR_ENCRYPTED_DATA_LEN_RANGE;
+        goto done;
     }
     // the amount of cleartext after stripping the padding will actually be less
     // than the input bytes...
@@ -315,13 +356,15 @@ CK_RV des_cbc_pad_decrypt(STDLL_TokData_t *tokdata,
 
     if (length_only == TRUE) {
         *out_data_len = padded_len;
-        return CKR_OK;
+        rc = CKR_OK;
+        goto done;
     }
 
     clear = (CK_BYTE *) malloc(padded_len);
     if (!clear) {
         TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
-        return CKR_HOST_MEMORY;
+        rc = CKR_HOST_MEMORY;
+        goto done;
     }
     rc = ckm_des_cbc_decrypt(tokdata, in_data, in_data_len, clear, &padded_len,
                              ctx->mech.pParameter, key);
@@ -333,6 +376,10 @@ CK_RV des_cbc_pad_decrypt(STDLL_TokData_t *tokdata,
 
     free(clear);
 
+done:
+    object_put(tokdata, key, TRUE);
+    key = NULL;
+
     return rc;
 }
 
@@ -379,7 +426,7 @@ CK_RV des_ecb_encrypt_update(STDLL_TokData_t *tokdata,
             return CKR_OK;
         }
 
-        rc = object_mgr_find_in_map_nocache(ctx->key, &key);
+        rc = object_mgr_find_in_map_nocache(tokdata, ctx->key, &key, READ_LOCK);
         if (rc != CKR_OK) {
             TRACE_ERROR("Failed to find specified object.\n");
             return rc;
@@ -388,6 +435,8 @@ CK_RV des_ecb_encrypt_update(STDLL_TokData_t *tokdata,
         clear = (CK_BYTE *) malloc(out_len);
         if (!clear) {
             TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
+            object_put(tokdata, key, TRUE);
+            key = NULL;
             return CKR_HOST_MEMORY;
         }
         // copy any data left over from the previous encryption operation first
@@ -407,7 +456,12 @@ CK_RV des_ecb_encrypt_update(STDLL_TokData_t *tokdata,
                 memcpy(context->data, in_data + (in_data_len - remain), remain);
             context->len = remain;
         }
+
         free(clear);
+
+        object_put(tokdata, key, TRUE);
+        key = NULL;
+
         return rc;
     }
 }
@@ -456,7 +510,7 @@ CK_RV des_ecb_decrypt_update(STDLL_TokData_t *tokdata,
             return CKR_OK;
         }
 
-        rc = object_mgr_find_in_map_nocache(ctx->key, &key);
+        rc = object_mgr_find_in_map_nocache(tokdata, ctx->key, &key, READ_LOCK);
         if (rc != CKR_OK) {
             TRACE_ERROR("Failed to find specified object.\n");
             return rc;
@@ -465,6 +519,8 @@ CK_RV des_ecb_decrypt_update(STDLL_TokData_t *tokdata,
         cipher = (CK_BYTE *) malloc(out_len);
         if (!cipher) {
             TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
+            object_put(tokdata, key, TRUE);
+            key = NULL;
             return CKR_HOST_MEMORY;
         }
         // copy any data left over from the previous decryption operation first
@@ -483,7 +539,12 @@ CK_RV des_ecb_decrypt_update(STDLL_TokData_t *tokdata,
                 memcpy(context->data, in_data + (in_data_len - remain), remain);
             context->len = remain;
         }
+
         free(cipher);
+
+        object_put(tokdata, key, TRUE);
+        key = NULL;
+
         return rc;
     }
 }
@@ -533,7 +594,7 @@ CK_RV des_cbc_encrypt_update(STDLL_TokData_t *tokdata,
             return CKR_OK;
         }
 
-        rc = object_mgr_find_in_map_nocache(ctx->key, &key);
+        rc = object_mgr_find_in_map_nocache(tokdata, ctx->key, &key, READ_LOCK);
         if (rc != CKR_OK) {
             TRACE_ERROR("Failed to find specified object.\n");
             return rc;
@@ -543,6 +604,8 @@ CK_RV des_cbc_encrypt_update(STDLL_TokData_t *tokdata,
         clear = (CK_BYTE *) malloc(out_len);
         if (!clear) {
             TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
+            object_put(tokdata, key, TRUE);
+            key = NULL;
             return CKR_HOST_MEMORY;
         }
         // copy any data left over from the previous encryption operation first
@@ -568,6 +631,10 @@ CK_RV des_cbc_encrypt_update(STDLL_TokData_t *tokdata,
         }
 
         free(clear);
+
+        object_put(tokdata, key, TRUE);
+        key = NULL;
+
         return rc;
     }
 }
@@ -617,7 +684,7 @@ CK_RV des_cbc_decrypt_update(STDLL_TokData_t *tokdata,
             return CKR_OK;
         }
 
-        rc = object_mgr_find_in_map_nocache(ctx->key, &key);
+        rc = object_mgr_find_in_map_nocache(tokdata, ctx->key, &key, READ_LOCK);
         if (rc != CKR_OK) {
             TRACE_ERROR("Failed to find specified object.\n");
             return rc;
@@ -627,6 +694,8 @@ CK_RV des_cbc_decrypt_update(STDLL_TokData_t *tokdata,
         cipher = (CK_BYTE *) malloc(out_len);
         if (!cipher) {
             TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
+            object_put(tokdata, key, TRUE);
+            key = NULL;
             return CKR_HOST_MEMORY;
         }
         // copy any data left over from the previous decryption operation first
@@ -652,6 +721,10 @@ CK_RV des_cbc_decrypt_update(STDLL_TokData_t *tokdata,
         }
 
         free(cipher);
+
+        object_put(tokdata, key, TRUE);
+        key = NULL;
+
         return rc;
     }
 }
@@ -685,7 +758,7 @@ CK_RV des_cbc_pad_encrypt_update(STDLL_TokData_t *tokdata,
     // note, this is subtly different from the other encrypt update routines
     //
     if (total <= DES_BLOCK_SIZE) {
-        if (length_only == FALSE) {
+        if (length_only == FALSE && in_data_len) {
             memcpy(context->data + context->len, in_data, in_data_len);
             context->len += in_data_len;
         }
@@ -711,7 +784,7 @@ CK_RV des_cbc_pad_encrypt_update(STDLL_TokData_t *tokdata,
         //    1) remain != 0
         //    2) out_len != 0
         //
-        rc = object_mgr_find_in_map_nocache(ctx->key, &key);
+        rc = object_mgr_find_in_map_nocache(tokdata, ctx->key, &key, READ_LOCK);
         if (rc != CKR_OK) {
             TRACE_ERROR("Failed to find specified object.\n");
             return rc;
@@ -721,6 +794,8 @@ CK_RV des_cbc_pad_encrypt_update(STDLL_TokData_t *tokdata,
         clear = (CK_BYTE *) malloc(out_len);
         if (!clear) {
             TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
+            object_put(tokdata, key, TRUE);
+            key = NULL;
             return CKR_HOST_MEMORY;
         }
         // copy any data left over from the previous encryption operation first
@@ -747,6 +822,10 @@ CK_RV des_cbc_pad_encrypt_update(STDLL_TokData_t *tokdata,
         }
 
         free(clear);
+
+        object_put(tokdata, key, TRUE);
+        key = NULL;
+
         return rc;
     }
 }
@@ -780,7 +859,7 @@ CK_RV des_cbc_pad_decrypt_update(STDLL_TokData_t *tokdata,
     // note, this is subtly different from the other decrypt update routines
     //
     if (total <= DES_BLOCK_SIZE) {
-        if (length_only == FALSE) {
+        if (length_only == FALSE && in_data_len) {
             memcpy(context->data + context->len, in_data, in_data_len);
             context->len += in_data_len;
         }
@@ -806,7 +885,7 @@ CK_RV des_cbc_pad_decrypt_update(STDLL_TokData_t *tokdata,
         //    1) remain != 0
         //    2) out_len != 0
         //
-        rc = object_mgr_find_in_map_nocache(ctx->key, &key);
+        rc = object_mgr_find_in_map_nocache(tokdata, ctx->key, &key, READ_LOCK);
         if (rc != CKR_OK) {
             TRACE_ERROR("Failed to find specified object.\n");
             return rc;
@@ -816,6 +895,8 @@ CK_RV des_cbc_pad_decrypt_update(STDLL_TokData_t *tokdata,
         cipher = (CK_BYTE *) malloc(out_len);
         if (!cipher) {
             TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
+            object_put(tokdata, key, TRUE);
+            key = NULL;
             return CKR_HOST_MEMORY;
         }
         // copy any data left over from the previous decryption operation first
@@ -838,6 +919,10 @@ CK_RV des_cbc_pad_decrypt_update(STDLL_TokData_t *tokdata,
             context->len = remain;
         }
         free(cipher);
+
+        object_put(tokdata, key, TRUE);
+        key = NULL;
+
         return rc;
     }
 }
@@ -1019,7 +1104,7 @@ CK_RV des_cbc_pad_encrypt_final(STDLL_TokData_t *tokdata,
         return CKR_FUNCTION_FAILED;
     }
 
-    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key);
+    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key, READ_LOCK);
     if (rc != CKR_OK) {
         TRACE_ERROR("Failed to find specified object.\n");
         return rc;
@@ -1040,7 +1125,7 @@ CK_RV des_cbc_pad_encrypt_final(STDLL_TokData_t *tokdata,
 
     if (length_only == TRUE) {
         *out_data_len = out_len;
-        return CKR_OK;
+        rc = CKR_OK;
     } else {
         memcpy(clear, context->data, context->len);
 
@@ -1049,8 +1134,12 @@ CK_RV des_cbc_pad_encrypt_final(STDLL_TokData_t *tokdata,
 
         rc = ckm_des_cbc_encrypt(tokdata, clear, out_len, out_data,
                                  out_data_len, ctx->mech.pParameter, key);
-        return rc;
     }
+
+    object_put(tokdata, key, TRUE);
+    key = NULL;
+
+    return rc;
 }
 
 
@@ -1073,7 +1162,7 @@ CK_RV des_cbc_pad_decrypt_final(STDLL_TokData_t *tokdata,
         TRACE_ERROR("%s received bad argument(s)\n", __func__);
         return CKR_FUNCTION_FAILED;
     }
-    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key);
+    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key, READ_LOCK);
     if (rc != CKR_OK) {
         TRACE_ERROR("Failed to find specified object.\n");
         return rc;
@@ -1085,7 +1174,8 @@ CK_RV des_cbc_pad_decrypt_final(STDLL_TokData_t *tokdata,
     //
     if (context->len != DES_BLOCK_SIZE) {
         TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_LEN_RANGE));
-        return CKR_ENCRYPTED_DATA_LEN_RANGE;
+        rc = CKR_ENCRYPTED_DATA_LEN_RANGE;
+        goto done;
     }
     // we don't know a priori how much data we'll be returning. we won't
     // know until after we decrypt it and strip the padding. it's possible
@@ -1095,7 +1185,7 @@ CK_RV des_cbc_pad_decrypt_final(STDLL_TokData_t *tokdata,
 
     if (length_only == TRUE) {
         *out_data_len = out_len;
-        return CKR_OK;
+        rc = CKR_OK;
     } else {
         memcpy(cipher, context->data, DES_BLOCK_SIZE);
 
@@ -1110,8 +1200,13 @@ CK_RV des_cbc_pad_decrypt_final(STDLL_TokData_t *tokdata,
             *out_data_len = out_len;
         }
 
-        return rc;
     }
+
+done:
+    object_put(tokdata, key, TRUE);
+    key = NULL;
+
+    return rc;
 }
 
 
@@ -1131,33 +1226,22 @@ CK_RV ckm_des_key_gen(STDLL_TokData_t *tokdata, TEMPLATE *tmpl)
     CK_ATTRIBUTE *class_attr = NULL;
     CK_ATTRIBUTE *local_attr = NULL;
     CK_BYTE *des_key = NULL;
-    CK_BYTE dummy_key[DES_KEY_SIZE] = { 0, };
     CK_ULONG rc;
-    CK_ULONG keysize;
+    CK_ULONG keysize = 0;
+    CK_BBOOL is_opaque = FALSE;
 
     if (token_specific.t_des_key_gen == NULL) {
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
         return CKR_MECHANISM_INVALID;
     }
 
-    if (is_secure_key_token())
-        keysize = token_specific.token_keysize;
-    else
-        keysize = DES_KEY_SIZE;
-
-    if ((des_key = (CK_BYTE *) calloc(1, keysize)) == NULL) {
-        TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
-        return CKR_HOST_MEMORY;
-    }
-
-    rc = token_specific.t_des_key_gen(tokdata, des_key, keysize, DES_KEY_SIZE);
+    rc = token_specific.t_des_key_gen(tokdata, &des_key, &keysize,
+                                      DES_KEY_SIZE, &is_opaque);
     if (rc != CKR_OK)
         goto err;
 
-    /* For secure-key keys put in CKA_IBM_OPAQUE
-     * and put dummy_key in CKA_VALUE.
-     */
-    if (is_secure_key_token()) {
+    /* For opaque keys put in CKA_IBM_OPAQUE and put dummy_key in CKA_VALUE. */
+    if (is_opaque) {
         opaque_attr = (CK_ATTRIBUTE *) malloc(sizeof(CK_ATTRIBUTE) + keysize);
         if (!opaque_attr) {
             TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
@@ -1168,7 +1252,18 @@ CK_RV ckm_des_key_gen(STDLL_TokData_t *tokdata, TEMPLATE *tmpl)
         opaque_attr->ulValueLen = keysize;
         opaque_attr->pValue = (CK_BYTE *) opaque_attr + sizeof(CK_ATTRIBUTE);
         memcpy(opaque_attr->pValue, des_key, keysize);
-        template_update_attribute(tmpl, opaque_attr);
+        rc = template_update_attribute(tmpl, opaque_attr);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("template_update_attribute failed\n");
+            free(opaque_attr);
+            goto err;
+        }
+    } else {
+        if (keysize != DES_KEY_SIZE) {
+            TRACE_ERROR("Invalid key size: %lu\n", keysize);
+            rc = CKR_FUNCTION_FAILED;
+            goto err;
+        }
     }
 
     value_attr = (CK_ATTRIBUTE *) malloc(sizeof(CK_ATTRIBUTE) + DES_KEY_SIZE);
@@ -1180,15 +1275,6 @@ CK_RV ckm_des_key_gen(STDLL_TokData_t *tokdata, TEMPLATE *tmpl)
         (CK_ATTRIBUTE *) malloc(sizeof(CK_ATTRIBUTE) + sizeof(CK_BBOOL));
 
     if (!value_attr || !key_type_attr || !class_attr || !local_attr) {
-        if (value_attr)
-            free(value_attr);
-        if (key_type_attr)
-            free(key_type_attr);
-        if (class_attr)
-            free(class_attr);
-        if (local_attr)
-            free(local_attr);
-
         TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
         rc = CKR_HOST_MEMORY;
         goto err;
@@ -1197,8 +1283,8 @@ CK_RV ckm_des_key_gen(STDLL_TokData_t *tokdata, TEMPLATE *tmpl)
     value_attr->type = CKA_VALUE;
     value_attr->ulValueLen = DES_KEY_SIZE;
     value_attr->pValue = (CK_BYTE *) value_attr + sizeof(CK_ATTRIBUTE);
-    if (is_secure_key_token())
-        memcpy(value_attr->pValue, dummy_key, DES_KEY_SIZE);
+    if (is_opaque)
+        memset(value_attr->pValue, 0, DES_KEY_SIZE);
     else
         memcpy(value_attr->pValue, des_key, DES_KEY_SIZE);
     free(des_key);
@@ -1218,16 +1304,44 @@ CK_RV ckm_des_key_gen(STDLL_TokData_t *tokdata, TEMPLATE *tmpl)
     local_attr->pValue = (CK_BYTE *) local_attr + sizeof(CK_ATTRIBUTE);
     *(CK_BBOOL *) local_attr->pValue = TRUE;
 
-    template_update_attribute(tmpl, value_attr);
-    template_update_attribute(tmpl, key_type_attr);
-    template_update_attribute(tmpl, class_attr);
-    template_update_attribute(tmpl, local_attr);
+    rc = template_update_attribute(tmpl, value_attr);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("template_update_attribute failed\n");
+        goto err;
+    }
+    value_attr = NULL;
+    rc = template_update_attribute(tmpl, key_type_attr);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("template_update_attribute failed\n");
+        goto err;
+    }
+    key_type_attr = NULL;
+    rc = template_update_attribute(tmpl, class_attr);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("template_update_attribute failed\n");
+        goto err;
+    }
+    class_attr = NULL;
+    rc = template_update_attribute(tmpl, local_attr);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("template_update_attribute failed\n");
+        goto err;
+    }
+    local_attr = NULL;
 
     return CKR_OK;
 
 err:
     if (des_key)
         free(des_key);
+    if (value_attr)
+        free(value_attr);
+    if (key_type_attr)
+        free(key_type_attr);
+    if (class_attr)
+        free(class_attr);
+    if (local_attr)
+        free(local_attr);
 
     return rc;
 }
@@ -1265,17 +1379,9 @@ CK_RV ckm_cdmf_key_gen(STDLL_TokData_t *tokdata, TEMPLATE *tmpl)
         (CK_ATTRIBUTE *) malloc(sizeof(CK_ATTRIBUTE) + sizeof(CK_BBOOL));
 
     if (!value_attr || !key_type_attr || !class_attr || !local_attr) {
-        if (value_attr)
-            free(value_attr);
-        if (key_type_attr)
-            free(key_type_attr);
-        if (class_attr)
-            free(class_attr);
-        if (local_attr)
-            free(local_attr);
-
         TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
-        return CKR_HOST_MEMORY;
+        rc = CKR_HOST_MEMORY;
+        goto error;
     }
 
     value_attr->type = CKA_VALUE;
@@ -1298,12 +1404,44 @@ CK_RV ckm_cdmf_key_gen(STDLL_TokData_t *tokdata, TEMPLATE *tmpl)
     local_attr->pValue = (CK_BYTE *) local_attr + sizeof(CK_ATTRIBUTE);
     *(CK_BBOOL *) local_attr->pValue = TRUE;
 
-    template_update_attribute(tmpl, value_attr);
-    template_update_attribute(tmpl, key_type_attr);
-    template_update_attribute(tmpl, class_attr);
-    template_update_attribute(tmpl, local_attr);
+    rc = template_update_attribute(tmpl, value_attr);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("template_update_attribute failed\n");
+        goto error;
+    }
+    value_attr = NULL;
+    rc = template_update_attribute(tmpl, key_type_attr);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("template_update_attribute failed\n");
+        goto error;
+    }
+    key_type_attr = NULL;
+    rc = template_update_attribute(tmpl, class_attr);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("template_update_attribute failed\n");
+        goto error;
+    }
+    class_attr = NULL;
+    rc = template_update_attribute(tmpl, local_attr);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("template_update_attribute failed\n");
+        goto error;
+    }
+    local_attr = NULL;
 
     return CKR_OK;
+
+error:
+    if (value_attr)
+        free(value_attr);
+    if (key_type_attr)
+        free(key_type_attr);
+    if (class_attr)
+        free(class_attr);
+    if (local_attr)
+        free(local_attr);
+
+    return rc;
 }
 
 #endif
